#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

#define BUFFER_SIZE 256

int checkEnd(FILE *file) {
	char c = fgetc(file);

	if (feof(file))
		return 1;
	
	ungetc(c, file);
	return 0;
}

void phraseNum(FILE *file) {
	int phrase_counter = 1;
	while (1) {
        if (checkEnd(file))
            break;

        char c = fgetc(file);
        if (c == '.' || c == '!' || c == '?') {
            if (checkEnd(file))
                break;
            else
                phrase_counter++;
        }
	}

	printf("%d\n", phrase_counter);
}

void clearBuffer(char *buffer) {
    for (int i = 0; i < BUFFER_SIZE; i++)
        buffer[i] = '\0'; // default value
}

void append(char* s, char c) {
    int len = strlen(s);

	if (len == BUFFER_SIZE - 1) {
		write(STDOUT_FILENO, s, strlen(s));
		clearBuffer(s);
		s[0] = c;
	}
	else {
		s[len] = c;
		s[len + 1] = '\0';
	}
}

void displayPhrases(FILE *file) {
	int phrase_counter = 1, process = 1;
	while (process) {
        char phrase[BUFFER_SIZE];

        char counter = phrase_counter + 48;
        append(phrase, '[');
        append(phrase, counter);
        append(phrase, ']');

		char c = fgetc(file);
		if (c != ' ')
			ungetc(c, file);
        append(phrase, ' ');
		
		while (1) {
			if (checkEnd(file)) {
				append(phrase, '\n');
				process = 0;
				break;
			}

			c = fgetc(file);

			if (c != '\n') {
				append(phrase, c);
				if (c == '.' || c == '!' || c == '?') {
					if (checkEnd(file))
						process = 0;
                    else
                        phrase_counter++;

					append(phrase, '\n');
					break;
				}
			}
		}

        write(STDOUT_FILENO, phrase, strlen(phrase));
        clearBuffer(phrase);
	}
}

int main(int argc, char* argv[]) {
	/* check arguments */
	if (!(argc == 2 || (argc == 3 && !strcmp(argv[1], "-l")))) {
		printf("usage: phrases [-l] file\n");
		return EXIT_FAILURE;
	}
	
	/* check if file can be opened and is readable */
	int file = argc - 1;
	FILE *fd = fopen(argv[file], "r");
	if (fd == NULL) {
		printf("error: cannot open %s\n", argv[file]);
		return EXIT_FAILURE;
	}
	
	/* check if file is empty */
	if (checkEnd(fd)) {
		printf("file is empty\n");
		return EXIT_FAILURE;
	}

	if (argc == 2)
		phraseNum(fd); // show number of phrases
	else if (argc == 3)	
		displayPhrases(fd); // print the contents
	
	/* close file */
	fclose(fd);

    return EXIT_SUCCESS;
}
