#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/wait.h>

int fgeti(FILE *stream) {
	return fgetc(stream) - 48;
}

void getDimensions(FILE *file, int *rows, int *cols) {
	*rows = fgeti(file);
	fgetc(file); // to ignore 'x'
	*cols = fgeti(file);
	fgetc(file); // to ignore '\n'
}

int main(int argc, char *argv[]) {	
	if (argc != 3){
		printf("usage: addmx file1 file2\n");
		return -1;
	}

	FILE *fd1, *fd2;
	fd1 = fopen(argv[1], "r");
	fd2 = fopen(argv[2], "r");

	if (fd1 == NULL) {
		printf ("error: cannot open %s\n", argv[1]);
		exit(1);
	}
	else if (fd2 == NULL) {
		printf ("error: cannot open %s\n", argv[2]);
		exit(1);
	}
	
	int rows1, cols1, rows2, cols2;
	getDimensions(fd1, &rows1, &cols1);
	getDimensions(fd2, &rows2, &cols2);

	if ((rows1 != rows2) || (cols1 != cols2)) {
		printf("dimensions don't match\n");
		return EXIT_FAILURE;
	}

    int rows = rows1, cols = cols1;

    /* setup shared memory */

    size_t matrix_size = rows * cols * sizeof(int);
    int *m = mmap(NULL, 3 * matrix_size, PROT_READ|PROT_WRITE, MAP_SHARED|MAP_ANONYMOUS, 0, 0);

    if(m == MAP_FAILED){
        perror("mmap");
        exit(EXIT_FAILURE);
    }

    /* read 1st matrix */
    for (int i = 0; i < rows1 * cols1; i++) {
        char c1 = fgetc(fd1);
		if (c1 == ' ' || c1 == '\n')
			m[i] = fgeti(fd1);
		else {
            ungetc(c1, fd1);
			m[i] = fgeti(fd1);
        }
    }

    /* read 2nd matrix */
    for (int i = 0; i < rows2 * cols2; i++) {
        char c2 = fgetc(fd2);
		if (c2 == ' ' || c2 == '\n')
			m[rows1 * cols1 + i] = fgeti(fd2);
		else {
            ungetc(c2, fd2);
			m[rows1 * cols1 + i] = fgeti(fd2);
        }
    }

	fclose(fd1);
	fclose(fd2);

    /* start nprocs and do work */
    int nprocs = cols;
    for(int j = 0; j < nprocs; j++) {
        pid_t pid;
        if ((pid = fork()) < 0) {
            perror("fork");
            exit(EXIT_FAILURE);
        }
        if(pid == 0) {
            for (int i = 0; i < rows; i++)
                m[(2*rows*cols) + (i*cols + j)] = m[i*cols + j] + m[(rows*cols) + (i*cols + j)];

            exit(EXIT_SUCCESS);
        }
    }

    /* wait for nprocs to finish */
    for(int i = 0; i < nprocs; i++) {
        if (waitpid(-1, NULL, 0) < 0) {
            perror("waitpid");
            exit(EXIT_FAILURE);
        }
    }

    /* read results sent by the child processes */
    printf("%dx%d\n", rows, cols);
    for (int i = 0; i < rows; i++) {
        for (int j = 0; j < cols; j++)
            printf("%d ", m[(2*rows*cols) + (i*cols + j)]);
        printf("\n");
    }

    /* release shared memory */
    if (munmap(m, sizeof(m)) < 0) {
        perror("munmap");
        exit(EXIT_FAILURE);
    }
	
	return EXIT_SUCCESS;
}
