#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdbool.h>
#include <sys/wait.h>
#include <errno.h>

#define READ_END 0
#define WRITE_END 1

#define LINESIZE 256
#define MAX_WORDS 30
#define MAX_WORD_SIZE 50
#define BUFFER_SIZE 1024

char* replaceWords(const char* text, const char* left, const char* right) {
    char* result;
    int l, lcnt = 0, r, rcnt = 0;
    int rightlen = strlen(right);
    int leftlen = strlen(left);
  
    for (l = 0; text[l] != '\0'; l++) {
        if (strstr(&text[l], left) == &text[l]) {
            lcnt++;
            l += leftlen - 1;
        }
    }

    for (r = 0; text[r] != '\0'; r++) {
        if (strstr(&text[r], right) == &text[r]) {
            rcnt++;
            r += rightlen - 1;
        }
    }
  
    result = (char*)malloc(l + lcnt * (rightlen - leftlen) + rcnt * (leftlen - rightlen) + 1);
  
    int i = 0;
    while (*text) {
        if (strstr(text, left) == text) {
            strcpy(&result[i], right);
            i += rightlen;
            text += leftlen;
        }
        else if (strstr(text, right) == text) {
            strcpy(&result[i], left);
            i += leftlen;
            text += rightlen;
        }
        else
            result[i++] = *text++;
    }
  
    result[i] = '\0';
    return result;
}

char* cypherText(char *text) {
    /* check cypher file */
    FILE *c;
    if ((c = fopen("cypher.txt", "r")) == NULL){
        fprintf(stderr, "error: cannot open %s\n", "cypher.txt");
        exit(EXIT_FAILURE);
    }

    char *result = (char*)malloc(strlen(text));
    strcpy(result, text);
    
    char left[MAX_WORD_SIZE];
    char right[MAX_WORD_SIZE];

    size_t c_size = 0;
    char* c_words = NULL;
    
    while (getline(&c_words, &c_size, c) > 0) {
        char *token = strtok(c_words, " ");
        strcpy(left, token);

        token = strtok(NULL, "\n");
        strcpy(right, token);

        result = replaceWords(result, left, right);
    }
    
    fclose(c);

    return result;
}

void clearBuffer(char *buffer) {
    for (int i = 0; i < BUFFER_SIZE; i++)
        buffer[i] = '\0'; // default value
}

int main(int argc, char* argv[]) {
    /* set pipes */

    int fd1[2]; // parent -> child
    if (pipe(fd1) < 0) {
        perror("pipe error");
        exit(EXIT_FAILURE);
    }

    int fd2[2]; // child -> parent
    if (pipe(fd2) < 0) {
        perror("pipe error");
        exit(EXIT_FAILURE);
    }
    
    /* start processes and do work */

    pid_t pid;

    if ((pid = fork()) < 0) {
        perror("fork error");
        exit(EXIT_FAILURE);
    }
    else if (pid > 0) { // parent
        close(fd2[WRITE_END]);
        close(fd1[READ_END]);

        /* read input text */

        char* content = NULL;
        size_t size = 0;

        char* line = NULL;
        size_t s = LINESIZE;

        while (getline(&line, &s, stdin) > 0) {
            size += strlen(line);
            content = (char*)realloc(content, size * sizeof(char) + 1);
            strcat(content, line);
        }

        /* write message to child */
        if (write(fd1[WRITE_END], content, strlen(content)) < 0)
            fprintf(stderr, "Unable to write to pipe: %s\n", strerror(errno));

        free(content);

        /* read message from child */
        char buffer[BUFFER_SIZE];
        int nbytes;
        while (true) {
            if ((nbytes = read(fd2[READ_END], buffer, BUFFER_SIZE)) < 0)
                fprintf(stderr, "Unable to read from pipe: %s\n", strerror(errno));

            write(STDOUT_FILENO, buffer, nbytes);

            if (nbytes < BUFFER_SIZE)
                break;
            else
                clearBuffer(buffer);
        }

        close(fd2[READ_END]);
        close(fd1[WRITE_END]);

        /* wait for child */
        if (waitpid(pid, NULL, 0) < 0) {
            fprintf(stderr, "Cannot wait for child: %s\n", strerror(errno));
        }
        
        exit(EXIT_SUCCESS);
    }
    else { // child
        close(fd2[READ_END]);
        close(fd1[WRITE_END]);

        /* read message from parent */

        char* content = NULL;
        size_t size = 0;

        char buffer[BUFFER_SIZE];
        clearBuffer(buffer);
        int nbytes;
        while (true) {
            if ((nbytes = read(fd1[READ_END], buffer, BUFFER_SIZE - 1)) < 0)
                fprintf(stderr, "Unable to read from pipe: %s\n", strerror(errno));
  
            size += strlen(buffer);
            content = (char*)realloc(content, size * sizeof(char) + 1);
            strcat(content, buffer);

            if (strlen(buffer) < BUFFER_SIZE - 1)
                break;
            else
                clearBuffer(buffer);
        }

        char *result = cypherText(content); // swap words

        /* write message to parent */
        if (write(fd2[WRITE_END], result, strlen(result)) < 0)
            fprintf(stderr, "Unable to write to pipe: %s\n", strerror(errno));

        free(result);
        free(content);
        
        close(fd2[WRITE_END]);
        close(fd1[READ_END]);

        /* exit gracefully */
        exit(EXIT_SUCCESS);
    }

    return EXIT_FAILURE;
}
